﻿#nullable disable

using System.Data;
using System.Reflection;

namespace AdoNetWrapper.Transaction.Common;

/// <summary>
/// Abstract base class for all 
/// ADO.NET database operations
/// </summary>
public abstract partial class DatabaseContext : IDisposable {
  public DatabaseContext(string connectString) {
    ConnectionString = connectString;
    Init();
  }

  public string ConnectionString { get; set; }
  public string ParameterPrefix { get; set; }
  public bool IsInTransaction { get; set; }
  public IDbConnection ConnectionObject { get; set; }
  public IDbTransaction TransactionObject { get; set; }
  public IDbCommand CommandObject { get; set; }
  public IDataReader DataReaderObject { get; set; }

  protected virtual void Init() {
    ParameterPrefix = string.Empty;
  }

  public virtual IDbConnection CreateConnection() {
    return CreateConnection(ConnectionString);
  }

  public abstract IDbConnection CreateConnection(string connectString);

  public virtual IDbCommand CreateCommand(string sql) {
    return CreateCommand(CreateConnection(), sql);
  }

  public abstract IDbCommand CreateCommand(IDbConnection cnn, string sql);

  public abstract IDataParameter CreateParameter(string paramName, object value);

  public abstract IDataParameter CreateParameter();

  public abstract IDataParameter GetParameter(string paramName);

  public virtual IDataReader CreateDataReader() {
    return CreateDataReader(CommandObject, CommandBehavior.CloseConnection);
  }

  public virtual IDataReader CreateDataReader(CommandBehavior cmdBehavior) {
    return CreateDataReader(CommandObject, cmdBehavior);
  }

  public virtual IDataReader CreateDataReader(IDbCommand cmd, CommandBehavior cmdBehavior = CommandBehavior.CloseConnection) {
    // Open Connection
    cmd.Connection.Open();

    // Create DataReader
    DataReaderObject = cmd.ExecuteReader(cmdBehavior);

    return DataReaderObject;
  }

  public abstract void GetLastAutoIncrement<TEntity>(IDbCommand cmd, TEntity entity, PropertyInfo prop);

  public virtual void BeginTransaction() {
    // Create Connection
    ConnectionObject = CreateConnection();
    // Open Connection
    ConnectionObject.Open();
    // Create Transaction
    TransactionObject = ConnectionObject.BeginTransaction();
    IsInTransaction = true;
  }

  public virtual void CommitTransaction() {
    if (ConnectionObject != null && ConnectionObject.State == ConnectionState.Open) {
      if (TransactionObject != null) {
        TransactionObject.Commit();
        ConnectionObject.Close();
        IsInTransaction = false;
      }
    }
  }

  public virtual void RollbackTransaction() {
    if (ConnectionObject != null && ConnectionObject.State == ConnectionState.Open) {
      if (TransactionObject != null) {
        TransactionObject.Rollback();
        ConnectionObject.Close();
        IsInTransaction = false;
      }
    }
  }

  public virtual void Dispose() {
    // Close/Dispose of data reader object
    if (DataReaderObject != null) {
      DataReaderObject.Close();
      DataReaderObject.Dispose();
    }

    // Close/Dispose of command object
    if (CommandObject != null) {
      if (CommandObject.Connection != null) {
        if (CommandObject.Transaction != null) {
          CommandObject.Transaction.Dispose();
        }
        CommandObject.Connection.Close();
        CommandObject.Connection.Dispose();
      }
      CommandObject.Dispose();
    }

    // Close/Dispose of Transaction object
    if (TransactionObject != null) {
      TransactionObject.Dispose();
    }

    // Close/Dispose of Connection object
    if (ConnectionObject != null) {
      ConnectionObject.Close();
      ConnectionObject.Dispose();
    }
  }
}
